#include <torch/extension.h>
#include <torch/torch.h>
#include <vector>

// A combined and optimized LSTM forward kernel
torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training) {

    // Ensure h0 and c0 are on the same device as x
    auto device = x.device();
    h0 = h0.to(device);
    c0 = c0.to(device);

    auto out = x;
    auto hn = h0.clone();
    auto cn = c0.clone();

    const size_t num_layers = lstm_weights_ih.size();

    // Lambda to process a single LSTM layer
    auto process_layer = [&](size_t i) {
        // Get the weights and biases for this layer
        auto weight_ih = lstm_weights_ih[i];
        auto weight_hh = lstm_weights_hh[i];
        auto bias_ih = lstm_biases_ih[i];
        auto bias_hh = lstm_biases_hh[i];

        int64_t input_size = weight_ih.size(1);
        int64_t hidden_size = weight_hh.size(1);

        // Instantiate a single-layer LSTM model
        torch::nn::LSTM lstm_model(
            torch::nn::LSTMOptions(input_size, hidden_size)
                .num_layers(1)
                .batch_first(true)
                .bidirectional(false));
        lstm_model->to(device);

        // Copy weights and biases using a pragma unroll to help compile-time unrolling
        #pragma unroll
        {
            lstm_model->named_parameters()["weight_ih_l0"].copy_(weight_ih);
            lstm_model->named_parameters()["weight_hh_l0"].copy_(weight_hh);
            lstm_model->named_parameters()["bias_ih_l0"].copy_(bias_ih);
            lstm_model->named_parameters()["bias_hh_l0"].copy_(bias_hh);
        }

        // Get the corresponding hidden and cell state slice for this layer
        auto h_slice = hn.narrow(0, i, 1);
        auto c_slice = cn.narrow(0, i, 1);
        std::tuple<torch::Tensor, torch::Tensor> state_tuple = std::make_tuple(h_slice, c_slice);

        // Set training mode accordingly
        lstm_model->train(is_training);

        // Forward pass through the current LSTM layer
        auto output_and_state = lstm_model->forward(out, state_tuple);
        auto output = std::get<0>(output_and_state);
        auto h_n_c_n = std::get<1>(output_and_state);

        // Update hidden and cell state slices
        #pragma unroll
        {
            auto h_n = std::get<0>(h_n_c_n);
            auto c_n = std::get<1>(h_n_c_n);
            hn.narrow(0, i, 1).copy_(h_n);
            cn.narrow(0, i, 1).copy_(c_n);
        }
        
        // Update output for the next layer
        out = output;
    };

    // Explicitly unroll the first 4 layers for common cases, then loop for remaining layers
    if (num_layers > 0) process_layer(0);
    if (num_layers > 1) process_layer(1);
    if (num_layers > 2) process_layer(2);
    if (num_layers > 3) process_layer(3);
    for (size_t i = 4; i < num_layers; ++i) {
        process_layer(i);
    }

    return hn;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Optimized LSTM forward (CUDA)");
}
